 #!/usr/bin/perl
package ChunkAlignment;
use base "Exporter";
@EXPORT = qw(new);

#use String::Similarity;
use Lingua::EN::Tagger;
use WordNet::QueryData;

use WordNet::Similarity::vector;
use WordNet::SenseRelate::AllWords;
use List::Util qw( min max );
use Text::Levenshtein;

use strict;
use warnings;


#standardizes tokens (makes lowercase, removes spaces, removes punctuation, etc...)
#input - a token (string of characters)
#ouptut - the input token standardized
sub standardizeToken
{
    my $token = shift;

    #make the token all lowercase
    $token = lc $token;

    #remove any spaces
    $token =~ s/ //g;

    #make 's into is
    if ($token eq "\'s") {
	$token = 'is';
    }

    #remove any punctuation
    $token =~ s/[.,;'?!"]//g;

    #TODO look for hyphenated words like "building-collapse" ...these should be split into 2 words but then how do I work indexing?

    return $token;
}


#standardizes each token in the list of tokens (makes lowercase, removes spaces, removes punctuation, etc...)
#input - a list of tokens
#ouptput the list of tokens standardized
sub standardizeTokens
{
    my @tokens = @{ shift @_ };
    #standardize each token
    for (my $i = 0; $i < scalar @tokens; $i++) {
	$tokens[$i] = &standardizeToken($tokens[$i]);
    }
    return @tokens;
}


#create the chunk data structures
#   for each word in the sentence there is a number indicating which chunk it belongs to
#   if no alignment then the number is always 0   #TODO huh?  so punctuation is always 0?
#input - a sentence with chunks seperated by square brackets (e.g. [ the boy ] [ and ] [ the girl ] )
#output - a list of numbers that indicate the chunk number of each word in the sentence (e.g. 0 0 1 2 2 )
sub createChunkDataStructures
{
    #grabe the input
    my @sentence = @{ shift @_ };

    #create a chunk mapping for each sentence
    my @chunks = ();
    my $count = 0;
    my $chunkNumber = 0;
    foreach my $chunk(@sentence) {
	$chunk =~ s/[\[\]]//g;  #remove brackets
	my @words = split(" ", $chunk);  #break into words
	foreach my $word(@words) {
	    $chunks[$count] = $chunkNumber;
	    $count++;
	 }
	 $chunkNumber++;
    }

    #return the chunk mapping
    return @chunks;
}

#tokenizes a raw input sentence (remove brackets and builds an array)
#input - a pre-chunked sentence
#output - an array  of tokens with each element a token in the original sentence
sub tokenize
{
    my @rawSentence = @{ shift @_ }; 

    #remove the square brakets([ , ]) and put into an array of words
    my $joined = join(" ",@rawSentence);
    $joined =~ tr/[]//d;
    my @tokens = split(" ",$joined);

    #return the token list
    return @tokens;
}

#gets part of speech (POS)  tags for a sentence
#input - a sentence or phrase
#output - an array of POS tags corresponding to each word in the input
sub getPOSTags
{
    #grab the input
    my $sentence = shift;
    my $tagger = shift;

    #Add POS Tags to the text
    my $taggedText1 = $tagger->add_tags($sentence);
    my %wordList1 = $tagger->get_words($sentence);
    my $tempText = $tagger->get_readable($sentence);
    my @taggedText = split(/ /,$tempText);

    #break pos tags off into a seperate array
    my @posTags = ();
    for(my $i=0; $i< scalar @taggedText; $i++) {
	my @wordAndTags= split(/\//,$taggedText[$i]); #break the tag off
	$posTags[$i]=$wordAndTags[1];  #1 is the pos tag index
    }

    #return the posTags
    return @posTags;
}


#performs word sense disambiguation on a list of tokens
#input - an array of tokens and a parallel array of POS tags.  Also an instance of SenseRelate::AllWords
#returns - a list of WordNet Synsets that is parallel to the input array of words.  if no match was found an empty string is placed in the array
sub disambiguateTokens
{
    #------ Align Disambiguated Terms with POS tagged Terms -------
    #The Disambiguation will leave out some terms, join others, and mark some as not defined
    #   here we are matching the disambiguated terms w/in wordnet to the input indeces
    
    #grab input
    my @words = @{ shift @_ };
    my @posTags = @{ shift @_ };
    my $wsd = shift;

    #Disambiguate Tokens
    my @disambiguated = $wsd->disambiguate (window => scalar @words, context => [@words]);
    my @wordNetTerms = ();
    for (my $i = 0; $i < scalar @words; $i++) {
	$wordNetTerms[$i] = "";  #begin by assuming the word is not in wordnet
	for (my $j = 0; $j < scalar @disambiguated; $j++) {

	    #check if the posTag is puncuation (can cause regex errors)
	    if ($posTags[$i] =~ m/(PP|PPC|PPD|PPL|PPS|LRB|RRR|SYM)/) {
		$wordNetTerms[$i] = "";
	    }
	    else {  #it is not puncuation
		#see if the disambiguated term contains the word
		#   and make sure it is the full word (not just a part of the word like (an and banner)
		if ($disambiguated[$j] =~ m/$words[$i]#/   #case where word matches
		    || $disambiguated[$j] =~ m/$words[$i]_/  #case where word is a compound word
		    || $disambiguated[$j] =~ m/_$words[$i]/ ) {  #case where word is a compound word

		    #there was a match, add it to the list of wordNetTerms only if its defined (not #ND)
		    if ($disambiguated[$j] !~ m/#ND/
			&& $disambiguated[$j] !~ m/#NR/) {
			$wordNetTerms[$i] = $disambiguated[$j];
		    }
		    last; #Match found, move to the next word
		}
	    }
	}
    }

    #return the wordnet terms list
    return @wordNetTerms;
}



#Computes the similarity between two words. Uses a WordNet Measure for words in WordNet 
#   or a modified the Levenstein distance for words not in WordNet
sub computeWordToWordSimilarity
{
    #grab the input
    my $token1 = shift;
    my $token2 = shift;
    my $wordNetTerm1 = shift;
    my $wordNetTerm2 = shift;
    my $measure = shift;
    my $wq = shift;
    my $wordToWordSimilarityThreshold = shift; 
    my $maxLevenshtein = shift;  

    #get all WordNet senses of the words (so it works with partially or fully qualified terms)
    my @senses1 = $wq->getSenses($wordNetTerm1);
    my @senses2 = $wq->getSenses($wordNetTerm2);

    #only use similarity measures if both terms are in WordNet
    if (scalar @senses1 > 0 && scalar @senses2 > 0) {
	#one or both of the terms are in WordNet, compute their similarity/relatedness

	my $maxSimilarity = 0;
	foreach my $sense1(@senses1) {
	    foreach my $sense2(@senses2) {
		#determine the relatednessclear
		my $similarity = $measure->getRelatedness($sense1, $sense2);

		#threshold similarity to account for rounding error
		if ($similarity < 0) {
		    $similarity = 0;
		}
		elsif ($similarity > 1) {
		    $similarity = 1;
		}

		#check for errors
		my ($error, $errorString) = $measure->getError();
		if ($error) {
		    print "senses1 = $sense1\n";
		    print "senses2 = $sense2\n";
		    die $errorString;
		}

		#update maxSimilarity
		if ($similarity > $maxSimilarity) {
		    $maxSimilarity = $similarity;
		}
	    }
	}
    		    
        #threshold similarity 
        if ($maxSimilarity < $wordToWordSimilarityThreshold) {
	    $maxSimilarity = 0;
        }

        #update the similarity score
        return $maxSimilarity
    }
    else {
	#one or both of the terms are not in WordNet, compute their levenshtein distance
	#make any empty strings or punctuation be an alignment of 0
	if ($token1 eq "" || $token2 eq "") {
	    return 0;
	}
	if ($token1 =~ /[,.!\?]/ || $token2 =~ /[,.!?]/) {
	    return 0;
	}

	#determine similarity based on Levenshtein distance
	if ($maxLevenshtein == 0) {
	    #words match only if they are identical
	    if ($token1 eq $token2) {
		return 1;
	    }
	    return 0;
	}
	else {
	    #compute the Levenshtein Distance
	    my $distance = &Text::Levenshtein::distance($token1, $token2);

	    #threshold, scale and return the distance
	    if ($distance > $maxLevenshtein) {
		return 0;
	    }
	    else {
		return ($maxLevenshtein - $distance)/$maxLevenshtein;
	    }
	}
    }
}



#computes the similairity between all words in the input arrays
#   returns a matrix of similarity scores
sub computeWordToWordSimilarities
{
    #grab input
    my @tokens1 = @{ shift @_ };
    my @tokens2 = @{ shift @_ };
    my @wordNetTerms1 = @{ shift @_ };
    my @wordNetTerms2 = @{ shift @_ };
    my $measure = shift;   
    my $wq = shift;
    my $wordToWordSimilarityThreshold = shift;
    my $maxLevenshtein = shift;

    #Initialize Similarities Data Structure, a matrix of Word-to-Word Similarities
    my @similarities = ();
    for (my $i = 0; $i < scalar @tokens1; $i++) {
	my @termSimilarities = ();
	$similarities[$i] = \@termSimilarities;
    }

    #Compute the similarity between all words
    for (my $i = 0; $i < scalar @wordNetTerms1; $i++) {
	my @termSimilarities = @{ $similarities[$i] };

	#compare the term against each term in the other sentence
	for (my $j = 0; $j < scalar @wordNetTerms2; $j++) {
	    $termSimilarities[$j] = &computeWordToWordSimilarity($tokens1[$i], $tokens2[$j], 
								 $wordNetTerms1[$i], $wordNetTerms2[$j], 
								 $measure, $wq,
								 $wordToWordSimilarityThreshold, $maxLevenshtein);
	}
    	#update that row of the matrix
 	$similarities[$i] = \@termSimilarities;
    }

    #return the word-to-word similarities
    return \@similarities;
}


#removes the stop words from a list of tokens using the POS tags
sub removeStopWords
{
    #grab input
    my @tokens = @{ shift @_ };
    my @posTags = @{ shift @_ };

    #define regex constants
    my $POS_Noun = "((NN)|(NNP)|(NNPS)|(NNS))";  #Non-pronoun nouns
    my $POS_Verb = "((VB)|(VBD)|(VBG)|(VBN)|(VBP)|(VBZ))";  #verbs (all tenses and forms)
    my $POS_Adjective = "((JJ)|(JJR)|(JJS)|(CD))"; #adjectives and cardinal numbers
    my $POS_Adverb = "((RB)|(RBR)|(RBS)|(RP))";  #adverbs

    #remove any words not tagged as a noun, verb, adjective, or adverb
    for (my $i = 0; $i < scalar @posTags; $i++) {
	my $posTag = $posTags[$i];
	
	#only nouns, verbs, adjectives, and adverbs are in wordnet
	if (!($posTag =~ m/$POS_Noun/
	      || $posTag =~ m/$POS_Verb/
	      || $posTag =~ m/$POS_Adjective/
	      || $posTag =~ m/$POS_Adverb/)) {
	    #the posTag didn't match any of the non-stop word types, so remove the word
	    $tokens[$i] = '';
	}
    }

    #return the stop-word-removed tokens
    return @tokens;
}



#computes the similarity between 2 chunks
sub computeChunkSimilarity
{
    #grab input
    my @chunks1 = @{ shift @_ };
    my @chunks2 = @{ shift @_ };
    my @similarities = @{ shift @_ };
    my $chunk1Number = shift;
    my $chunk2Number = shift;
    my $chunkSimilarityThreshold = shift;

    #compute the similarity between the two chunks
    my $chunk1Length = 0;
    my $chunk2Length = 0;
    my $totalSimilarity = 0;
    for(my $i = 0; $i < scalar @chunks1; $i++) {

	if ($chunks1[$i] == $chunk1Number) {
	    #this word is a part of chunk1, compare it to all words in current chunk2
	    $chunk1Length++;
	    my @wordSimilarities = @ { $similarities[$i] };

	    #find the max similarity
	    $chunk2Length = 0;
	    my $maxSimilarity = 0;
	    for (my $j = 0; $j < scalar @chunks2; $j++) {
		if ($chunks2[$j] == $chunk2Number) {
			  
		    #this word is part of chunk2, grab the similarity and update the maximum similarity
		    $chunk2Length++;
		    if ($wordSimilarities[$j] > $maxSimilarity) {
			$maxSimilarity = $wordSimilarities[$j];
		     }
		 }
	    }	

	    #update this chunk's total similarity with the max similarity for that word
	    $totalSimilarity += $maxSimilarity;

       } #end if part of chunk1 if
    } #end chunks21 loop
	   
    #normalize the total similarity by the length of the shorter chunk
    if ($chunk1Length > $chunk2Length) {
	if ($chunk2Length != 0) {
	    $totalSimilarity /= $chunk2Length;
	}
    } else {
	if ($chunk1Length != 0) {
	    $totalSimilarity /= $chunk1Length;
	}
    }
	    
    #apply a threshold
    if ($totalSimilarity < $chunkSimilarityThreshold) {
	$totalSimilarity = 0;
    }

    #return the total similarity between the two chunks
    return $totalSimilarity;
}


#aligns the chunked data of two sentences, finds the alignmnet scores, and posTags
sub alignChunks
{
    #----- grab the input -----
    my @rawSentence1 = @{ shift @_ };
    my @rawSentence2 = @{ shift @_ };
    my $qd = shift;
    my $wntools = shift;
    my $wsd = shift;
    my $measure = shift;
    my $tagger = shift;
    my $wq = shift;
    my $wordToWordSimilarityThreshold = shift;
    my $chunkSimilarityThreshold = shift;
    my $maxLevenshtein = shift;
    my $useWSD = shift;
    my $removeStopWords = shift;
    my $rechunk = shift;

    #tokenize the sentences
    my @rawTokens1 = &tokenize(\@rawSentence1);
    my @rawTokens2 = &tokenize(\@rawSentence2);

    #standardize the tokens
    my @tokens1 = &standardizeTokens(\@rawTokens1);
    my @tokens2 = &standardizeTokens(\@rawTokens2);

    #Get POS tags for each token
    my @posTags1 = &getPOSTags(join(" ", @rawTokens1), $tagger);
    my @posTags2 = &getPOSTags(join(" ", @rawTokens2), $tagger);

    #Disambiguate Tokens
    my @wordNetTerms1;
    my @wordNetTerms2;
    if ($useWSD > 0) {
	@wordNetTerms1 = &disambiguateTokens(\@tokens1, \@posTags1, $wsd);
	@wordNetTerms2 = &disambiguateTokens(\@tokens2, \@posTags2, $wsd);
    } else {
	@wordNetTerms1 = @tokens1;
	@wordNetTerms2 = @tokens2;
    }

    #remove stop words
    if ($removeStopWords > 0) {
	@tokens1 = &removeStopWords(\@tokens1, \@posTags1);
	@tokens2 = &removeStopWords(\@tokens2, \@posTags2);
    }
    
    #compute word to word similarities for all chunks
    my @similarities = @{ 
	&computeWordToWordSimilarities(\@tokens1, \@tokens2,  
				       \@wordNetTerms1, \@wordNetTerms2, 
				       $measure, $wq, 
				       $wordToWordSimilarityThreshold, $maxLevenshtein) };

    #add to the list of word similarities
    foreach my $similarityRef(@similarities) {
    	my @wordSimilarities = @ { $similarityRef };
    }

    #create chunk arrays
    my @chunks1 = &createChunkDataStructures(\@rawSentence1);
    my @chunks2 = &createChunkDataStructures(\@rawSentence2);
    my $numChunks1 = (max @chunks1) + 1; #add one since the chunks are indeces
    my $numChunks2 = (max @chunks2) + 1;

    #compute the chunk to chunk similarities and chunk alignments
    my @chunkAlignments = ();
    my @chunkAlignmentSimilarities = ();
    for (my $i = 0; $i < $numChunks1; $i++) {
	
	#find the similarities between this chunk and every other chunk
	my @chunkSimilarities = ();
	my $maxSimilarity = 0;
	my $maxSimilarityIndex = -1;
	for (my $j = 0; $j < $numChunks2; $j++) {
	    $chunkSimilarities[$j] = 
		&computeChunkSimilarity(\@chunks1, \@chunks2, 
					\@similarities, $i, $j, $chunkSimilarityThreshold);

	    if ($chunkSimilarities[$j] > $maxSimilarity) {
		$maxSimilarity = $chunkSimilarities[$j];
		$maxSimilarityIndex = $j;
	    }
	}

	#set the chunk alignment
	if ($maxSimilarity < $chunkSimilarityThreshold) {
	    #there was no alignment
	    $chunkAlignments[$i] = 0;
	}
	else {
	    #there was an alignment
	    $chunkAlignments[$i] = $maxSimilarityIndex + 1; #add one since no alignment is 0
	}
	$chunkAlignmentSimilarities[$i] = $chunkSimilarities[$maxSimilarityIndex];
    }

    #rechunk (combine chunks that are aligned to the same thing) 
    if ($rechunk > 0) {
	my @results = &rechunk(\@chunks1, \@chunks2, 
			       \@chunkAlignments, \@chunkAlignmentSimilarities, 
			       \@similarities, $chunkSimilarityThreshold);
	@chunks1 = @{ $results[0] };
	@chunkAlignments = @{ $results[1] };
	@chunkAlignmentSimilarities = @{ $results[2] };

    }
    
    #return the results
    return (\@tokens1, \@tokens2, 
	    \@chunks1, \@chunks2, 
	    \@posTags1, \@posTags2, 
	    \@chunkAlignments, \@chunkAlignmentSimilarities);

}


#rechunks the data so that any chunks1 that are aligned to the same chunks2 become a single chunk
#   e.g.   "[cat] [and] [dog] <=> [cat dog]" becomes "[cat dog] [and] <=> [cat dog]"
#          @chunks = (0 1 2) <=> (0 0) @alignments = (1 0 1)... rechunked @chunks = (0 1 0) @alignments = (1 0)
sub rechunk
{
    my @chunks1 = @{ shift @_ };
    my @chunks2 = @{ shift @_ };
    my @alignments = @{ shift @_ };
    my @chunkSimilarities = @{ shift @_ };
    my @tokenSimilarities = @{ shift @_ };
    my $chunkSimilarityThreshold = shift;

    #go through each alignment number and see if another chunk is aligned with the same thing
    my $rechunkedCount = 0;
    for (my $i = 0; $i < scalar @alignments; $i++) {
	#skip no alignments
	if ($alignments[$i] > 0) {
	    #check every other chunk
	    for (my $j = ($i+1); $j < scalar @alignments; $j++) {

		#see if the two chunks1 are alignned with the same chunks2
		if ($alignments[$i] == $alignments[$j]) {
		    #match found, set all chunk numbers of this chunk to be the other chunk
		    for (my $k = 0; $k < scalar @chunks1; $k++) {
			if ($chunks1[$k] == $j) {
			    $chunks1[$k] = $i;
			}
		    }
		    #recompute the chunk similarity
		    $chunkSimilarities[$i] =  
			&computeChunkSimilarity(\@chunks1, \@chunks2, 
						\@tokenSimilarities, 
						$i, $alignments[$i], 
						$chunkSimilarityThreshold);

		}
	    }
	}
    }

    #create the rechunked data structures
    my @rechunkedAlignments = ();
    my @rechunkedSimilarities = ();
    
    #loop through the chunked data and renumber chunks
    my $count = 0;
    for (my $i = 0; $i < scalar @chunks1; $i++) {  #start at 1 since NOALI is always added to 0

	#relabel any chunk number $i's to chunk number $count
	my $exists = 0;
	for (my $j = 0; $j < scalar @chunks1; $j++) {
	    if ($chunks1[$j] == $i) {
		$exists = 1;
		$chunks1[$j] = $count;
	    }
	}
	#update the chunk count
	if ($exists) {
	    $rechunkedAlignments[$count] = $alignments[$i];
	    $rechunkedSimilarities[$count] = $chunkSimilarities[$i];
	    $count++;
	}
    }

    return \@chunks1, \@rechunkedAlignments, \@rechunkedSimilarities;
}


1;
